import logging
from difflib import SequenceMatcher # For title similarity
from langchain_openai import ChatOpenAI
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import StrOutputParser, JsonOutputParser
from langchain.agents import AgentExecutor, create_openai_functions_agent

from .db_core import get_system_prompt, get_app_settings
from .db_agent_ops import save_final_article, add_processed_url, is_url_processed
from .tools import collector_tools # Import prepared tool lists
from .agent_helpers import parse_json_output, title_similarity # Import helpers
from .llm_setup import initialize_llm # Import from correct location

logger = logging.getLogger(__name__)

# Old initialize_llm and parse_json_output removed, now imported from llm_setup and agent_helpers

# --- Collector Agent ---

def run_collector_agent(tags: list[str]):
    """
    Runs the Collector agent to find new articles for the given tags.
    Args:
        tags: A list of tags to search for.
    Returns:
        A list of dictionaries, where each dictionary represents a successfully scraped new article
        (e.g., {'url': '...', 'title': '...', 'content': '...'}).
        Returns an empty list if no new articles are found or an error occurs.
    """
    logger.info(f"Collector Agent: Starting run for tags: {tags}")
    try:
        llm, token_callback = initialize_llm(provider='openai') # Get LLM and callback
        prompt_str = get_system_prompt('Collector')
        if not prompt_str:
            logger.error("Collector Agent: Failed to load system prompt.")
            return []

        # Define the prompt template - Adjust based on expected input/output
        # This is a simplified example; a more robust agent might need message history, etc.
        # Instructing JSON output directly in the prompt.
        prompt = ChatPromptTemplate.from_messages([
            ("system", prompt_str),
            ("human", "Find relevant news articles for the following tags: {tags_query}. Use the available tools: 'google_search', 'check_if_url_is_processed', 'scrape_web_content'. Follow the instructions precisely, especially regarding checking processed URLs BEFORE scraping and returning the final result ONLY as a JSON list."),
            # ("placeholder", "{agent_scratchpad}"),
        ])

        # Create the agent using Langchain's recommended method
        collector_agent = create_openai_functions_agent(llm, collector_tools, prompt)

        # Create the AgentExecutor
        agent_executor = AgentExecutor(
            agent=collector_agent,
            tools=collector_tools,
            verbose=True,
            callbacks=[token_callback], # Pass callback handler here
            handle_parsing_errors=True # Gracefully handle potential LLM output parsing issues
        )

        tags_query = ", ".join(tags)
        # Invoke the agent - the specific input format depends on the prompt template
        result = agent_executor.invoke({"tags_query": tags_query})

        # --- Output Parsing (Expecting JSON) ---
        # The 'output' structure depends heavily on the LLM's response and the agent type.
        # For create_openai_functions_agent, the result might be in result['output'].
        raw_output = result.get('output', '')

        # Use the helper to parse the JSON output
        parsed_articles = parse_json_output(raw_output, expected_keys=["url", "title", "content"], agent_name="Collector")

        # Ensure it's a list, default to empty list if parsing fails or returns wrong type
        articles_data = parsed_articles if isinstance(parsed_articles, list) else []

        logger.info(f"Collector Agent: Finished run for tags {tags}. Found {len(articles_data)} potential new articles.")
        return articles_data # Return the parsed list

    except Exception as e:
        logger.error(f"Collector Agent: Error during execution for tags {tags}: {e}", exc_info=True)
        return []

# --- Editor Agent ---

def run_editor_agent(articles_to_merge: list[dict]):
    """
    Runs the Editor agent to merge and summarize overlapping articles.
    Args:
        articles_to_merge: A list of article dictionaries (containing 'url', 'title', 'content').
    Returns:
        A dictionary containing the merged 'title' and 'summary', or None if error.
    """
    logger.info(f"Editor Agent: Starting run to merge {len(articles_to_merge)} articles.")
    if not articles_to_merge:
        logger.warning("Editor Agent: No articles provided for merging.")
        return None

    try:
        # Fetch settings to determine which LLM provider to use
        app_settings = get_app_settings()
        editor_provider = app_settings.get('editor_llm_provider', 'openai') # Default to openai
        llm, token_callback = initialize_llm(provider=editor_provider) # Initialize LLM based on setting
        prompt_str = get_system_prompt('Editor')
        if not prompt_str:
            logger.error("Editor Agent: Failed to load system prompt.")
            return None

        # Prepare input string from articles
        input_text = "Please merge and summarize the following articles:\n\n"
        for i, article in enumerate(articles_to_merge):
            input_text += f"Article {i+1}:\n"
            input_text += f"  Title: {article.get('title', 'N/A')}\n"
            input_text += f"  URL: {article.get('url', 'N/A')}\n"
            # Include only the start of the content to avoid overly long prompts if necessary
            content_preview = article.get('content', 'N/A')[:1000] # Limit content length in prompt
            input_text += f"  Content (first 1000 chars): {content_preview}...\n\n"

        # Use a chain with JSON output parser
        prompt = ChatPromptTemplate.from_messages([
            ("system", prompt_str),
            ("human", f"{input_text}\n\nGenerate the merged title and summary in JSON format as instructed.")
        ])
        # Expecting JSON output based on the updated system prompt
        chain = prompt | llm | JsonOutputParser()

        # Invoke the chain, passing the callback handler
        # Input is already in the prompt object, but invoke needs a dict
        result = chain.invoke({}, config={"callbacks": [token_callback]})

        # --- Output Parsing ---
        # Assume the LLM returns the summary directly. Might need JSON parsing if structured output is requested.
        # Example: Parse "Title: [title]\nSummary: [summary]"
        # parsed_result = parse_editor_output(result) # Implement this function
        parsed_result = {"title": "Merged Article", "summary": result} # Placeholder

        # The JsonOutputParser should return a dict directly if successful
        if isinstance(result, dict) and "merged_title" in result and "merged_summary" in result:
            parsed_result = {"title": result["merged_title"], "summary": result["merged_summary"]}
            logger.info(f"Editor Agent: Finished merging. Result title: {parsed_result.get('title')}")
        else:
            logger.error(f"Editor Agent: Failed to parse expected JSON output. Raw result: {result}")
            return None
        return parsed_result

    except Exception as e:
        logger.error(f"Editor Agent: Error during execution: {e}", exc_info=True)
        return None
